{
  "type": "record",
  "Aspect": {
    "name": "structuredPropertyKey",
    "keyForEntity": "structuredProperty",
    "entityCategory": "core",
    "entityAspects": [
      "propertyDefinition",
      "structuredPropertySettings",
      "institutionalMemory",
      "status"
    ],
    "entityDoc": "Structured Property represents a property meant for extending the core model of a logical entity"
  },
  "name": "StructuredPropertyKey",
  "namespace": "com.linkedin.pegasus2avro.structured",
  "fields": [
    {
      "type": "string",
      "name": "id",
      "doc": "The id for a structured proeprty."
    }
  ]
}